#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from dataclasses import dataclass

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from httpx import Response

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from aioipfabric.base_client import IPFBaseClient, table_api
from aioipfabric.consts import COLOR_GREEN, TableFields

# -----------------------------------------------------------------------------
#
#                                 CODE BEGINS
#
# -----------------------------------------------------------------------------


@dataclass
class URIs:
    """ identifies API URL endpoings used"""

    devices = "/tables/inventory/devices"
    device_parts = "/tables/inventory/pn"
    managed_ipaddrs = "/tables/addressing/managed-devs"
    snapshots = "/snapshots"


DEFAULT_PARTS_COLUMNS = [
    "deviceSn",
    "hostname",
    "siteName",
    "name",
    "dscr",
    "pid",
    "sn",
    "vid",
    "vendor",
    "platform",
    "model",
]


class IPFInventoryMixin(IPFBaseClient):
    """
    IP Fabric client mixin supporting the inventory features.
        - devices
        - device optics
        - device managed IP addresses
    """

    @table_api
    async def fetch_devices(self, request: dict) -> Response:
        """
        Fetch <Inventory | Devices> table records.

        Parameters
        ----------
        request: dict
            The API body request payload, prepared by the table_api decorator

        Returns
        -------
        The HTTPx response, which will be post-processed by the table_api decorator.
        """

        default_columns = [
            "sn",
            "hostname",
            "siteKey",
            "siteName",
            "loginIp",
            "loginType",
            "uptime",
            "vendor",
            "platform",
            "family",
            "version",
            "model",
        ]
        request.setdefault(TableFields.columns, default_columns)
        return await self.api.post(URIs.devices, json=request)

    @table_api
    async def fetch_ipaddrs(self, request: dict) -> Response:
        """
        Fetch <Technology | Adressing | Managed IPs> table records.

        Parameters
        ----------
        request: dict
            The API body request payload, prepared by the table_api decorator

        Returns
        -------
        The HTTPx response, which will be post-processed by the table_api decorator.
        """
        default_columns = ["sn", "hostname", "intName", "siteName", "mac", "ip", "net"]

        request.setdefault("columns", default_columns)

        return await self.api.post(URIs.managed_ipaddrs, json=request)

    @table_api
    async def fetch_optics(self, request: dict) -> Response:
        """
        Fetch <Technology | Part numbers | SFP modules> table records.

        Parameters
        ----------
        request: dict
            The API body request payload, prepared by the table_api decorator

        Returns
        -------
        The HTTPx response, which will be post-processed by the table_api decorator.
        """
        filter_report = {"pid": ["color", "eq", COLOR_GREEN]}

        request.setdefault("columns", DEFAULT_PARTS_COLUMNS)
        request["filters"].update(filter_report)
        request["reports"] = "/inventory/part-numbers"

        return await self.api.post(URIs.device_parts, json=request)

    @table_api
    async def fetch_device_parts(self, request: dict) -> Response:
        """
        Fetch <Technology | Part numbers> table records.

        Parameters
        ----------
        request: dict
            The API body request payload, prepared by the table_api decorator

        Returns
        -------
        The HTTPx response, which will be post-processed by the table_api decorator.
        """
        request.setdefault("columns", DEFAULT_PARTS_COLUMNS)
        return await self.api.post(URIs.device_parts, json=request)



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import List, Dict, Optional, AnyStr
import csv

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from aioipfabric.mixins.inventory import IPFInventoryMixin

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["IPFabricClient"]


# -----------------------------------------------------------------------------
#
#                                 CODE BEGINS
#
# -----------------------------------------------------------------------------


class IPFabricClient(IPFInventoryMixin):
    """
    An instance IPFabricClient is used to interact with the IP Fabric
    system API via methods that abstract the underlying API calls.

    The IPFabricClient is composed of mixins, each of which address a different
    aspect of the IP Fabric product.  This composition structure allows the
    Developer to define a client with only those feature aspects that they need
    for their program.

    To dynamically add a Mixin class, use the `mixin` method defined
    by `IPFBaseClass`

    Examples
    --------
        from aioipfabric import IPFabricClient
        from aioipfabric.mixin_configs import IPFConfigMixin

        ipf = IPFabricClient()
        ipf.mixin(IPFConfigMixin)

    """

    @staticmethod
    def to_csv(
        datalist: List[Dict],
        filepath: AnyStr,
        fieldnames: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ) -> None:
        """
        This method will store the given list of dict items to a CSV file.  The
        CSV column headers will be taken from the first list item keys.  The
        `exclude` list can be used to omit designated columns from the CSV file,
        for exampel ['id'] would omit the 'id' column.

        Parameters
        ----------
        datalist
        filepath
        exclude
        fieldnames
        """
        dl_fieldnames = datalist[0].keys()

        if fieldnames:
            _s_fieldnames = set(fieldnames)
            _s_dl_fieldnames = set(dl_fieldnames)

            if _s_fieldnames & _s_dl_fieldnames != _s_fieldnames:
                raise RuntimeError(f"Invalid set of fieldnames: {fieldnames}")

            exclude = _s_dl_fieldnames - _s_fieldnames
        else:
            fieldnames = dl_fieldnames

        if exclude:
            fieldnames = [col for col in fieldnames if col not in exclude]
            for rec in datalist:
                for key in exclude:
                    del rec[key]

        with open(filepath, "w+") as ofile:
            csv_wr = csv.DictWriter(ofile, fieldnames=fieldnames)
            csv_wr.writeheader()
            csv_wr.writerows(datalist)



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional, Callable, Dict, Awaitable, List
from asyncio import Semaphore
from dataclasses import dataclass
import logging

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

import aioipfabric
from aioipfabric.aiofut import as_completed
from aioipfabric.base_client import IPFBaseClient

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["IPFConfigsMixin"]


# -----------------------------------------------------------------------------
#
#                                 CODE BEGINS
#
# -----------------------------------------------------------------------------

_LOG = logging.getLogger(aioipfabric.__package__)


@dataclass
class URIs:
    """ API endpoints """

    device_config_refs = "tables/management/configuration"
    download_device_config = "tables/management/configuration/download"


class IPFConfigsMixin(IPFBaseClient):
    """
    Mixin to support fetching device configurations (text)
    """

    async def fetch_device_configs(
        self,
        on_config: Callable[[Dict, str], Awaitable[None]],
        since_ts: int,
        all_configs: Optional[bool] = False,
        before_ts: Optional[int] = None,
        filters: Optional[Dict] = None,
        device_filter: Optional[Callable[[Dict], bool]] = None,
        sanitized: Optional[bool] = False,
        batch_sz: Optional[int] = 1,
        dry_run: Optional[bool] = False,
    ) -> List[Dict]:
        """
        This coroutine is used to download the latest copy of devices in the
        active snapshot.  The default behavior is to locate the lastest device
        configuration since the `since_ts` time for all devices.

        If the Caller needs to "go back in time" then both the `since_ts` and
        `before_ts` values should be used and only devices that match:

                since_ts <= lastChange <= before_ts

        If the Caller would like to further filter records based on 'hostname',
        'sn' (serial-number), those filters can be provided in the `filters`
        parameter.

        Parameters
        ----------
        since_ts:
            The timestamp criteria for retrieving configs such that lastChecked
            is >= since_ts. This value is the epoch timestamp * 1_000; which is
            the IP Fabric native storage unit for timestamps.

        all_configs:
            When False (default) this coroutine will fetech only those
            configurations that have _changed_.  Use all_configs=True when you
            want to fetch the configs regardless if they have changed or not.

        on_config:
            A coroutine that will be invoked with the device record and
            configuration file content that allows the Caller to do
            something with the content such as save it to a file.

        filters:
            Any additional Caller provided API filters that should be applied in
            addition to using the `since_ts` and `before_ts` options.  This
            filters dictionary allows the caller to be more specific, for
            example regex of hostnames.  The filters only apply to the fields to
            extact the backup record hashs; which primarily include ['sn',
            'hostname', 'status']

        device_filter:
            The Caller can optionally provide a function used to filter if the
            provided device-hash record should be used or not to retrieve the
            device configuration.  This parameter is useful for cases where the
            Caller wants to limit the device configuration retrival based on
            filtering beyond the get-hash filter fields of sn and hostname.

        before_ts:
            The timestamp criteria for retrieving configs such that lastChecked
            is <= before_ts.  This value is the epoch timestamp * 1_000.

        sanitized:
            Determines if the configuration should be santized as it is extracted
            from the IP Fabric system.

        batch_sz:
            Number of concurrent download tasks; used to rate-limit IPF API
            due to potential performance related issue.

        dry_run:
            When True this coroutine will return the list of devices
            that would be used to retrieve the configuration files; but
            not actually get the configs.

        Returns
        -------
        List of device-hash records that were used to perform the config file
        fetching process.

        Notes
        -----
        From experiments with the IP Fabric API, v3.6.1, observations are that
        batch_sz must be 1 as higher values result in result text as
        combinations of multiple devices. (Confirmed by IP Fabric engineering
        2020-Sep-09)
        """

        since_criteria = "lastCheck" if all_configs else "lastChange"

        # The first step is to retrieve each of the configuration "hash" records
        # using the active snapshot start timestamp as the basis for the filter.

        if before_ts:
            # TODO: The attempt to use the 'and' with this API call causes an
            #       Error 500 response code.  Leaving this code in for now so we
            #       can debug/troubleshoot with IPF team.

            if before_ts <= since_ts:
                raise ValueError(f"before_ts {before_ts} <= since_ts {since_ts}")

            filters_ = {since_criteria: ["gte", since_ts]}

            # filters_ = {
            #     "and": [
            #         {"lastCheck": ["gte", since_ts]},
            #         {"lastCheck": ["lte", before_ts]},
            #     ]
            # }

        else:
            filters_ = {since_criteria: ["gte", since_ts]}

        # if the Caller provided additional filters, add them now.

        if filters:
            filters_.update(filters)

        payload = {
            "columns": [
                "_id",
                "sn",
                "hostname",
                "lastChange",
                "lastCheck",
                "status",
                "hash",
            ],
            "filters": filters_,
            "snapshot": self.active_snapshot,
            "sort": {"column": since_criteria, "order": "desc"},
            "reports": "/management/configuration/first",
        }

        res = await self.api.post(URIs.device_config_refs, json=payload)
        res.raise_for_status()
        records = res.json()["data"]

        # NOTE: For unknown reasons, there are devices that have more than one
        # record in this response collection.  Therefore we need to retain only
        # the most recent value using a dict and the setdefault method

        if before_ts:
            records = [rec for rec in records if rec[since_criteria] <= before_ts]

        filtered_records = dict()

        for rec in records:
            filtered_records.setdefault(rec["sn"], rec)

        records = list(filtered_records.values())

        if dry_run is True:
            return records

        # TODO NOTE: API workaournd for v3.6
        # since we cannot use the API for before_ts, we need to perform a post
        # API filtering process now

        # TODO NOTE: API limitation
        # Now we need to retrieve each config file based on each record hash. We
        # will create a list of tasks for this process and run them concurrently
        # in batches (due to API related reasons).

        batching_sem = Semaphore(batch_sz)

        async def fetch_device_config(_hash):
            """ perform a config fetch limited by semaphore """
            async with batching_sem:
                api_res = await self.api.get(
                    URIs.download_device_config,
                    params={"hash": _hash, "sanitized": sanitized},
                    timeout=60,
                )
                return api_res

        # create a lookup dictionary that will map the fetch coroutine to the
        # hash record so that when the coroutine completes we can obtain the
        # source device record; this is needed per the use of the `as_completed`
        # function.

        device_filter = device_filter or (lambda x: True)

        fetch_tasks = {
            fetch_device_config(rec["hash"]): rec
            for rec in records
            if device_filter(rec)
        }

        _LOG.debug(
            f"Fetching {len(fetch_tasks)} device configurations in {batch_sz} batches ... "
        )

        async for task in as_completed(fetch_tasks, timeout=5 * 60):
            # the `task` instance will provide both the config text as a result,
            # and allow use to use the associated coroutine as a lookup so we
            # can obtain the device record associated with the config.

            rec = fetch_tasks[task.get_coro()]
            t_result = task.result()

            # pass the device record and device configuration back to the Caller
            # via the callback coroutine so that they can do what they want; for
            # example save the contents to filesystem.

            await on_config(rec, t_result.text)

        # return only the list of device records that were subject to backup
        # processing.

        return list(fetch_tasks.values())



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
This file contains API helpers for filtering data.  At the time of v3.6, the
online documentation can be found here:
https://docs.ipfabric.io/api/#header-filter-structure.  Unfortunately the
content is lacking many details.

Examples
--------
Simple Expressions:

    "hostname = foo"
    "hostname =~ 'abc.*'"

Grouped Expressions:
    Both site is equal to 'atl' and hostname contains 'sw2'
    and (site = atl, hostname ~ sw2, vendor = cisco)

    Either site is euqal to 'atl' or hostname contains 'sw2'
    or (site = atl, hostname ~ sw2)

Nested Group Expressions:
    or (
        and(site = atl, hostname ~ 'core'),
        and(site = chc, hostname =~ '.*club-switch2[12]')
    )


References
----------
    IP Fabric API docs:
    https://docs.ipfabric.io/api/#header-filter-structure.

    Parsimonious:
    https://github.com/erikrose/parsimonious

Notes
-----
Filter options for <string> type

    ["eq", <string | null>] - equal to a string
    ["neq", <string | null>] - not equal to a string
    ["like", <string | null>] - contain a string
    ["notlike", <string | null>] - not contain a string
    ["reg", <string | null>] - match a regular expression
    ["nreg", <string | null>] - not match a regular expression
    ["empty", <boolean>] - (not) empty

Filter options for <IP address> type

    ["cidr", <IPv4 in CIDR notation>] - fit in a range
    ["eq", <string | null>] - equal to a string
    ["neq", <string | null>] - not equal to a string
    ["like", <string | null>] - contain a string
    ["notlike", <string | null>] - not contain a string
    ["reg", <string | null>] - match a regular expression
    ["nreg", <string | null>] - not match a regular expression
    ["empty", <boolean>] - (not) empty

Filter options for <number> type

    ["eq", <number | null>] - equal to a number
    ["neq", <number | null>] - not equal to a number
    ["gt", <number | null>] - greater than a number
    ["gte", <number | null>] - greater than or equal to a number
    ["lt", <number | null>] - lower than a number
    ["lte", <number | null>] - lower than or equal to a number
    ["empty", <boolean>] - (not) empty

* Filter by "column" comparison
    For example, on the Inventory|Interfaces table, using this filter will compare the value
    of the "l1" column to the "l2" column:

    { "filters": {"l1":["column","eq","l2"]}
"""
# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from types import MappingProxyType
from itertools import chain

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from parsimonious import Grammar, NodeVisitor
from parsimonious.nodes import RegexNode

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["parse_filter"]

# -----------------------------------------------------------------------------
#
#                                 CODE BEGINS
#
# -----------------------------------------------------------------------------

_OPERATORS = MappingProxyType(
    {
        "=": "eq",  # equals exactly
        "!=": "neq",  # does not equal
        "~": "like",  # string contains
        "!~": "notlike",  # string does not contain
        "=~": "reg",  # match regular expression
        "!=~": "nreg",  # does not match regular expression
        "?": "empty",  # column is empty, provided rhs-value is either "true" or "false"
        "net": "cidr",  # value match using IP CIDR value
        "<": "lt",  # less than
        "<=": "lte",  # less than or equal to
        ">": "gt",  # greater than
        ">=": "gte",  # greather than or equal to
    }
)


FILTER_GRAMMER = r"""
#
# Expression parts
#
filter_expr         = group_expr / simple_expr
group_expr_list     = group_expr ws ("," ws group_expr)+
group_expr          = group_tok ws "(" ws expr_list ws ")"
expr_list           = simple_expr_list / group_expr_list
simple_expr_list    = simple_expr ws ("," ws simple_expr)+
simple_expr         = col_name ws (column_expr_rhs / color_expr_rhs / oper_expr_rhs)
num_oper_expr_rhs   = (num_oper / ei_oper) ws int_tok
str_oper_expr_rhs   = (str_oper / ei_oper) ws cmp_value_tok
oper_expr_rhs       = (num_oper_expr_rhs / str_oper_expr_rhs)
column_expr_rhs     = "column" ws ei_oper ws col_name
color_expr_rhs      = "color" ws num_oper_expr_rhs
#
# Token parts
#
col_name        = ~"[a-z0-9_\-]+"i
sq_words        = ~"[^']+"
dq_words        = ~"[^\"]+"
ws              = ~"\s*"
sq              = "'"
dq              = "\""
word            = ~r"[a-z0-9\.\/_\-]+"i
int_tok         = ~"\d+"
group_tok       = 'and' / 'or'
str_oper        = '!=~' / '=~' / 'net' / '!~' / '~' / '?'
num_oper        =  '<=' / '>=' / '<' / '>'
ei_oper         = '!=' / '='
cmp_value_tok   = sq_tok / dq_tok / word
sq_tok          = sq sq_words sq
dq_tok          = dq dq_words dq
"""


_grammer = Grammar(FILTER_GRAMMER)


class _FilterConstructor(NodeVisitor):
    """ parsimouneous node visitor for handlingn the FILTER_GRAMMER """

    def visit_group_expr(self, node, vc):  # noqa
        """ create a group_expr item """
        group_tok, _, _, _, filter_list, *_ = vc
        return {group_tok: filter_list}

    def visit_group_expr_list(self, node, vc):  # noqa
        """ create a list of group_expr items """
        expr_1, _, expr_n = vc
        expr_list = [
            expr_1,
            *(expr for expr in chain.from_iterable(expr_n) if isinstance(expr, dict)),
        ]
        return expr_list

    def visit_expr_list(self, node, vc):  # noqa
        """ create an expr_list item """
        return vc[0]

    def visit_simple_expr_list(self, node, vc):  # noqa
        """ return a list of filter dictionaries """
        expr_1, _, expr_n = vc
        return [
            expr_1,
            *(expr for expr in chain.from_iterable(expr_n) if isinstance(expr, dict)),
        ]

    def visit_simple_expr(self, node, vc):  # noqa
        """ return a filter dictionary """
        col, rhs = vc[0], vc[2][0]

        if rhs[0] == "empty":
            val = rhs[1]
            if (val := {"true": True, "false": False}.get(val.lower())) is None:
                raise RuntimeError("'empty' value must be either 'true' or 'false'")
            rhs[1] = val

        return {col: rhs}

    # -------------------------------------------------------------------------
    #               Right Hand Side (RHS) Expressions
    # -------------------------------------------------------------------------

    def visit_num_oper_expr_rhs(self, node, vc):  # noqa
        """ returns the number operator and RHS value """
        oper, _, tok = vc
        return [oper[0], tok]

    def visit_str_oper_expr_rhs(self, node, vc):  # noqa
        """ returns the string operator and RHS value """
        oper, _, tok = vc
        return [oper[0], tok]

    def visit_oper_expr_rhs(self, node, vc):  # noqa
        """ returns the operattor right-hand-side expression list item """
        return vc[0]

    def visit_column_expr_rhs(self, node, vc):  # noqa
        """ return the 'column' operation right-hand-side expression list item"""
        col_oper, _, oper, _, col_name = vc
        return [col_oper.text, oper, col_name]

    def visit_color_expr_rhs(self, node, vc):  # noqa
        """ return the 'column' operation + right-hand-side expression list item"""
        col_oper, _, oper_val = vc
        return [col_oper.text, *oper_val]

    # -------------------------------------------------------------------------
    #                      Token Expressions
    # -------------------------------------------------------------------------

    def visit_group_tok(self, node, vc):  # noqa
        """ returns the group operator (and, or) value """
        return node.text

    def visit_cmp_value_tok(self, node, vc):  # noqa
        """ children will either be a single node-value or a quoted-value """
        vc = vc.pop(0)

        if isinstance(vc, RegexNode):
            return vc.text

        value_node = vc[0] if len(vc) == 1 else vc[1]
        return value_node.text

    def visit_int_tok(self, node, vc):  # noqa
        """ returns the value as an integer """
        return int(node.text)

    def visit_col_name(self, node, vc):  # noqa
        """ returns the column name (str) """
        return node.text

    # -------------------------------------------------------------------------
    #                      Operator Nodes
    # -------------------------------------------------------------------------

    def visit_str_oper(self, node, vc):  # noqa
        """ returns the string operator in IPF API form """
        return _OPERATORS[node.text]

    def visit_num_oper(self, node, vc):  # noqa
        """ returns the number operator in IPF API form """
        return _OPERATORS[node.text]

    def visit_ei_oper(self, node, vc):  # noqa
        """ returns the equality operator in IPF API form """
        return _OPERATORS[node.text]

    def generic_visit(self, node, visited_children):
        """ pass through for nodes not explicility visited """
        return visited_children or node


_filter_builder = _FilterConstructor()


def parse_filter(expr: str) -> dict:
    """
    This function is used to convert a filter expression, as a string in the form
    of FILTER_GRAMMER, and return the IPF filter dictionary that is consumed by
    the `filters` body parameters.

    Parameters
    ----------
    expr
        The filter expression, for example "hostname = switch1.dc1"
    """
    res = _grammer.parse(expr.strip().replace("\n", ""))
    return _filter_builder.visit(res)[0]



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional
from dataclasses import dataclass

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from httpx import AsyncClient

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["IPFSession"]


# -----------------------------------------------------------------------------
#
#                           CODE BEGINS
#
# -----------------------------------------------------------------------------


@dataclass
class URIs:
    """ identifies API URL endpoings used"""

    login = "auth/login"
    token_refresh = "auth/token"


class IPFSession(AsyncClient):
    """
    The IPFSession instance is the asyncio base client used to interact with the
    IP Fabric API via REST calls.  The primary feature of the IPFSession class
    is to handle the authentication via login credentials and tokens.

    An instance of this class will be created by the IPFBaseClient, which in
    turns makes the api accessbile to any IPFabricClient instances.
    """

    def __init__(self, base_url, token=None, username=None, password=None):
        """
        Initialize the asyncio client session to the IP Fabric API

        Parameters
        ----------
        base_url: str
            The base URL of the IP fabric system

        token: str
            The refresh token

        username: str
            The login user-name

        password: str
            The login password
        """
        super().__init__(base_url=base_url, verify=False)

        self.__refresh_token = token
        self.__access_token = None

        if all((username, password)):
            self.__init_auth = self.__auth_userpass(
                username=username, password=password
            )
        elif token:
            self.__init_auth = self.__refresh_access_token(token)
        else:
            raise RuntimeError("MISSING required token or (username, password)")

        self.headers["Content-Type"] = "application/json"

    async def authenticate(self):
        """
        This coroutine is used to authenticate to the IPF server and obtain an access
        token.  This coroutine can be used for both the initial login process as well
        as the token refresh process.
        """
        try:
            await self.__init_auth
        except RuntimeError:
            # A runtime error will occur if the __init_auth has already been awaited.
            await self.refresh_token()

    @property
    def token(self):
        """ return the Refresh Token for later use/storage """
        return self.__refresh_token

    async def refresh_token(self, token: Optional[str] = None):
        """ using the refresh token, obtain a new access token """

        if token:
            self.__refresh_token = token

        assert self.__refresh_token is not None
        await self.__refresh_access_token(self.__refresh_token)
        self.headers["Authorization"] = f"Bearer {self.__access_token}"

    async def __refresh_access_token(self, refresh_token):
        """ underlying API call to update the access token """
        res = await self.post(URIs.token_refresh, json={"refreshToken": refresh_token})
        res.raise_for_status()
        body = res.json()
        self.__access_token = body["accessToken"]

    async def __auth_userpass(self, username, password):
        """ underlying API to call to authenticate using login credentials """
        res = await self.post(
            URIs.login, json={"username": username, "password": password}
        )
        res.raise_for_status()
        body = res.json()
        self.__access_token = body["accessToken"]
        self.__refresh_token = body["refreshToken"]
        self.headers["Authorization"] = f"Bearer {self.__access_token}"



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
This module contains IPF client mixins that perform the "diagram" queries.
"""

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional, Dict, Union
from dataclasses import dataclass

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from aioipfabric.base_client import IPFBaseClient


# -----------------------------------------------------------------------------
#
#                                 CODE BEGINS
#
# -----------------------------------------------------------------------------


@dataclass
class URIs:
    """ identifies API URL endpoints used"""

    end_to_end_path = "graph/end-to-end-path"


class IPFDiagramE2EMixin(IPFBaseClient):
    """ Mixin for End-to-End Path query """

    async def end_to_end_path(
        self,
        src_ip: str,
        dst_ip: Optional[str] = "0.0.0.0",
        proto: Optional[str] = "tcp",
        src_port: Optional[Union[str, int]] = 0,
        dst_port: Optional[Union[str, int]] = 0,
        check_rpf: Optional[bool] = False,
        check_asymmetric: Optional[bool] = False,
    ) -> Dict:
        """
        Execute an "End-to-End Path" diagram query for the given set of parameters.

        Parameters
        ----------
        src_ip
            Source IP address
        dst_ip
            Destination IP address
        proto
            Protocol: "tcp", "udp", or "icmp"
        src_port
            Source Port
        dst_port
            Destination Port
        check_rpf
            Boolean to check for reverse path forwarding
        check_asymmetric
            Boolean to check for asymmetric routine

        Returns
        -------
        Returns a dictionary with 'graph' and 'ad' primary keys.  For more details refer to this
        IPF blog: https://ipfabric.io/blog/end-to-end-path-simulation-with-api/
        """
        res = await self.api.get(
            URIs.end_to_end_path,
            params=dict(
                source=src_ip,
                sourcePort=src_port,
                destination=dst_ip,
                destinationPort=dst_port,
                protocol=proto,
                asymmetric=check_asymmetric,
                rpf=check_rpf,
                snapshot=self.active_snapshot,
            ),
        )
        res.raise_for_status()
        return res.json()



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from aioipfabric.client import IPFabricClient

from aioipfabric.base_client import (
    table_api,
    IPFBaseClient as IPFabricClientMixin,
)

__all__ = [IPFabricClient, IPFabricClientMixin, table_api]



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from setuptools import setup, find_packages

package_name = "aio-ipfabric"
package_version = open("VERSION").read().strip()


def requirements(filename="requirements.txt"):
    return open(filename.strip()).readlines()


with open("README.md", "r") as fh:
    long_description = fh.read()


# -----------------------------------------------------------------------------
#
#                                 Main Setup
#
# -----------------------------------------------------------------------------

setup(
    name=package_name,
    version=package_version,
    description="Python asyncio client for IP Fabric",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jeremy Schulman",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Networking",
    ],
)



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional, AnyStr, Iterable
from os import environ, getenv
from dataclasses import dataclass
from functools import wraps

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from .consts import ENV, API_VER, TableFields
from .api import IPFSession
from .filters import parse_filter

# -----------------------------------------------------------------------------
# Exports
# -----------------------------------------------------------------------------

__all__ = ["IPFBaseClient", "table_api"]


# -----------------------------------------------------------------------------
#
#                           CODE BEGINS
#
# -----------------------------------------------------------------------------


@dataclass
class URIs:
    """ identifies API URL endpoings used"""

    snapshots = "/snapshots"


def table_api(methcoro):
    """ Method decorator for all Table related APIs """

    @wraps(methcoro)
    async def wrapper(self, request=None, return_as="data", **kwargs):
        """ wrapper that prepares the API with default behaviors """

        payload = request or {}
        payload.setdefault(TableFields.snapshot, self.active_snapshot)
        payload.setdefault(TableFields.filters, kwargs.get(TableFields.filters) or {})

        if TableFields.columns in kwargs:
            payload[TableFields.columns] = kwargs[TableFields.columns]

        # TODO: perhaps add a default_pagination setting to the IP Client?
        #       for now the default will be no pagnication

        if TableFields.pagination in kwargs:
            payload["pagination"] = kwargs[TableFields.pagination]

        res = await methcoro(self, payload)

        if return_as == "raw":
            return res

        res.raise_for_status()
        body = res.json()

        return {"data": body["data"], "meta": body["_meta"], "body": body}[return_as]

    return wrapper


class IPFBaseClient(object):
    """
    The IPFabricClient instances is composed of one or more Mixins that are a
    subclass of IPFBaseClient.  Put another way, he IPFBaseClient provides the
    common code that is available to all Mixins.

    The primary purpose of the IPFBaseClass instance is to provide the `api`
    attribute, which is an instance of the IPFSession (async HTTP client
    instance).
    """

    @dataclass
    class ENV:
        """ identifies enviornment variables used """

        addr = "IPF_ADDR"
        username = "IPF_USERNAME"
        password = "IPF_PASSWORD"
        token = "IPF_TOKEN"

    def __init__(
        self,
        /,
        *mixin_classes,
        base_url: Optional[AnyStr] = None,
        token: Optional[AnyStr] = None,
        username: Optional[AnyStr] = None,
        password: Optional[AnyStr] = None,
    ):
        """
        Create an IP Fabric Client instance

        Parameters
        ----------
        base_url : str
            The IP Fabric system HTTPS URL, for example:
            https://my-ipfabric-server/

        username: str
            The IPF login user-name value

        password: str
            The IPF login password value

        token : str
            The IP Fabric Refresh Token that will be used to create the Access
            Token required by API calls.

        Notes
        -----
        The Caller can provide either the login credentials (username, password)
        or the refresh token.  One of these two are required.
        """

        token = token or getenv(ENV.token)
        base_url = base_url or environ[ENV.addr]
        username = username or getenv(ENV.username)
        password = password or getenv(ENV.password)

        # ensure that the base_url ends with a slash since we will be using
        # httpx with base_url. there is a known _requirement_ for
        # ends-with-slash which if not in place causes issues.
        # TODO: perhaps this issue has been fixed in latter versions of httpx; not
        #       seeing the same issue with httpx > 0.14
        # if not base_url.endswith("/"):
        #     base_url += "/"

        self.api = IPFSession(
            base_url=base_url + API_VER,
            token=token,
            username=username,
            password=password,
        )

        # dynamically add any Mixins at the time of client creation.  This
        # enables the caller to perform the mixin at runtime without having to
        # define a specific class.

        if mixin_classes:
            self.mixin(*mixin_classes)

        self.snapshots = None
        self.active_snapshot = None
        self.version = None  # the IPF product version

    async def login(self):
        """
        Coroutine to perform the initial login authentication process, retrieve the list
        of current snapshots, and set the `active_snapshot` attribute to the latest
        snapshot.
        """
        await self.api.authenticate()
        res = await self.api.get("/os/version")
        res.raise_for_status()
        self.version = res.json()["version"]
        await self.fetch_snapshots()

        self.active_snapshot = self.snapshots[0]["id"]

    async def logout(self):
        """ close the async connection """
        await self.api.aclose()

    async def fetch_snapshots(self) -> None:
        """ coroutine to retrieve all known snapshots, returns List[dict] records """
        res = await self.api.get(URIs.snapshots)
        res.raise_for_status()
        self.snapshots = res.json()

    def mixin(self, *mixin_cls):
        """
        This method allows the Caller to dynamically add a Mixin class
        to the existing IPF client instance.

        Parameters
        ----------
        mixin_cls: subclasses of IPFBaseClass
            The mixin classes whose methods will be added to the existing
            IPF client instance (self).

        References
        ----------
        https://stackoverflow.com/questions/8544983/dynamically-mixin-a-base-class-to-an-instance-in-python
        """
        self.__class__ = type(self.__class__.__name__, (self.__class__, *mixin_cls), {})

    def __repr__(self) -> Iterable[str]:
        """ override the default repr to show the IPF system base URL """
        cls_name = self.__class__.__name__
        base_url = self.api.base_url
        return f"{cls_name}: {base_url}"

    # -------------------------------------------------------------------------
    #                      ASYNC CONTEXT MANAGER METHODS
    # -------------------------------------------------------------------------

    async def __aenter__(self):
        """ login and return instance """
        await self.login()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """ close the http async api instance """
        await self.logout()

    # -------------------------------------------------------------------------
    #                             STATIC METHODS
    # -------------------------------------------------------------------------

    parse_filter = staticmethod(parse_filter)



#  Copyright 2020 Jeremy Schulman
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from dataclasses import dataclass
from enum import Enum

# -----------------------------------------------------------------------------
#
#                                    CODE BEGINS
#
# -----------------------------------------------------------------------------


@dataclass
class ENV:
    """ identifies enviornment variables used """

    addr = "IPF_ADDR"
    username = "IPF_USERNAME"
    password = "IPF_PASSWORD"
    token = "IPF_TOKEN"


API_VER = "/api/v1/"


COLOR_GREEN = 0
COLOR_BLUE = 10
COLOR_YELLOW = 20
COLOR_RED = 30


class TableFields(str, Enum):
    """ identifies the API Table request body fields """

    columns = "columns"
    filters = "filters"
    pagination = "pagination"
    snapshot = "snapshot"
    reports = "reports"



from typing import AsyncIterable, Iterable, Coroutine, Optional
import asyncio
from asyncio import Task

__all__ = ["as_completed"]


async def as_completed(
    aws: Iterable[Coroutine], timeout: Optional[int] = None
) -> AsyncIterable[Task]:
    """
    This async generator is used to "mimic" the behavior of the
    concurrent.futures.as_completed functionality. Usage of this as_completed
    generator is slightly different from the builtin asyncio version; see
    example below.

    The builtin asyncio.as_completed yields futures such that the originating
    coroutine can not be retrieved.  In order to obtain the originating
    coroutine these must be wrapped in futures as explained in this Stack
    Overflow: https://bit.ly/2AsPtJE

    Parameters
    ----------
    aws:
        An interable of coroutines that will be wrapped into futures and
        executed through the asyncio on_completed builtin.

    timeout: int
        (seconds, same as asyncio.as_completed):
        If provided an asyncio.TimeoutError will be raised if all of the
        coroutines have not completed within the timeout value.

    Yields
    ------
    asyncio.Task

    Examples:
    ---------

        # create a dictionary of key=coroutine, value=dict, where the value will
        # be used later when the coroutine completes

        tasks = {
            probe(rec.get('ipaddr') or rec.get('host')): rec
            for rec in inventory
        }

        async for probe_task in as_completed(tasks):
            try:
                # obtain the originating coroutine so we can use it as an index
                # into the tasks dictionary and obtain the associated inventory
                # record

                task_coro = probe_task.get_coro()
                rec = tasks[task_coro]

                # now obtain the coroutine return value using the `result`
                # method.

                probe_ok = 'OK' if probe_task.result() else 'FAIL'
                report[probe_ok].append(rec)

            except OSError as exc:
                probe_ok = 'ERROR'
                report['ERROR'].append((rec, exc))

            print(f"{rec['host']}: {probe_ok}")
    """
    loop = asyncio.get_running_loop()

    # The main gist is to "wrap" the coroutine into
    # "[futureW[futureO[coroutine]]]" where the outer futureW is what is handed
    # into the builtin asyncio.as_completed.  The inner futureO that wraps the
    # originating coroutine will call the futureW.set_result() when the
    # original futureO[coroutine] completes.  The call to set_result triggers
    # the event causing the futureW to be done, which is what results in the
    # builtin on_completed to yield the results.

    def wrap_coro(coro):
        """ used so we can retrieved the original coro via task.get_coro() """
        fut = asyncio.ensure_future(coro)
        wrapper = loop.create_future()
        fut.add_done_callback(wrapper.set_result)
        return wrapper

    for next_completed in asyncio.as_completed(
        [wrap_coro(coro) for coro in aws], timeout=timeout
    ):
        yield await next_completed
